@@ -42,6 +42,47 @@ extern gboolean GdmAllowRoot;
 extern gboolean GdmAllowRemoteRoot;
 extern gint GdmRetryDelay;
 
+static char *selected_user = NULL;
+
+void
+gdm_verify_select_user (const char *user)
+{
+	g_free (selected_user);
+	if (ve_string_empty (user))
+		selected_user = NULL;
+	else
+		selected_user = g_strdup (user);
+}
+
+static void
+print_cant_auth_errbox (void)
+{
+	gboolean is_capslock = FALSE;
+	const char *basemsg;
+	char *msg;
+	char *ret;
+
+	ret = gdm_slave_greeter_ctl (GDM_QUERY_CAPSLOCK, "");
+	if ( ! ve_string_empty (ret))
+		is_capslock = TRUE;
+	g_free (ret);
+
+	basemsg = _("\nIncorrect username or password.  "
+		    "Letters must be typed in the correct "
+		    "case.");
+	if (is_capslock) {
+		msg = g_strconcat (basemsg, "  ",
+				   _("Please make sure the "
+				     "Caps Lock key is not "
+				     "enabled."),
+				   NULL);
+	} else {
+		msg = g_strdup (basemsg);
+	}
+	gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX, msg);
+	g_free (msg);
+}
+
 /**
  * gdm_verify_user:
  * @username: Name of user or NULL if we should ask
@@ -60,27 +101,44 @@ gdm_verify_user (GdmDisplay *d,
 		 const gchar *display,
 		 gboolean local) 
 {
-    gchar *login, *passwd, *ppasswd, *auth_errmsg;
+    gchar *login, *passwd, *ppasswd;
     struct passwd *pwent;
 
     if (local)
 	    gdm_slave_greeter_ctl_no_ret (GDM_STARTTIMER, "");
 
-    /* Ask for the user's login */
     if (username == NULL) {
+authenticate_again:
+	    /* Ask for the user's login */
+	    gdm_verify_select_user (NULL);
 	    gdm_slave_greeter_ctl_no_ret (GDM_MSG, _("Please enter your username"));
-	    login = gdm_slave_greeter_ctl (GDM_LOGIN, _("Username:"));
+	    login = gdm_slave_greeter_ctl (GDM_PROMPT, _("Username:"));
 	    if (login == NULL ||
 		gdm_slave_greeter_check_interruption ()) {
-		    if (local)
-			    gdm_slave_greeter_ctl_no_ret (GDM_STOPTIMER, "");
-		    g_free (login);
-		    return NULL;
+		    if ( ! ve_string_empty (selected_user)) {
+			    /* user selected */
+			    g_free (login);
+			    login = selected_user;
+			    selected_user = NULL;
+		    } else {
+			    /* some other interruption */
+			    if (local)
+				    gdm_slave_greeter_ctl_no_ret (GDM_STOPTIMER, "");
+			    g_free (login);
+			    return NULL;
+		    }
 	    }
 	    gdm_slave_greeter_ctl_no_ret (GDM_MSG, "");
+
+	    if ( ! gdm_slave_check_user_wants_to_log_in (login)) {
+		    g_free (login);
+		    login = NULL;
+		    goto authenticate_again;
+	    }
     } else {
 	    login = g_strdup (username);
     }
+    gdm_slave_greeter_ctl_no_ret (GDM_SETLOGIN, login);
 
     pwent = getpwnam (login);
         
@@ -92,7 +150,7 @@ gdm_verify_user (GdmDisplay *d,
 	    /* eeek a passwordless account */
 	    passwd = g_strdup ("");
     } else {
-	    passwd = gdm_slave_greeter_ctl (GDM_NOECHO, _("Password: "));
+	    passwd = gdm_slave_greeter_ctl (GDM_NOECHO, _("Password:"));
 	    if (passwd == NULL)
 		    passwd = g_strdup ("");
 	    if (gdm_slave_greeter_check_interruption ()) {
@@ -110,17 +168,10 @@ gdm_verify_user (GdmDisplay *d,
 
     if (pwent == NULL) {
 	    sleep (GdmRetryDelay);
-	    gdm_error (_("Couldn't authenticate user"));
-	    /* FIXME: Hmm, how are we sure that the login is username
-	     * and password.  That is the most common case but not
-	     * neccessairly true, this message needs to be changed
-	     * to allow for such cases */
-	    auth_errmsg = g_strdup_printf
-		    (_("\nIncorrect username or password.  "
-		       "Letters must be typed in the correct case.  "  
-		       "Please be sure the Caps Lock key is not enabled"));
-	    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX, auth_errmsg);
-	    g_free (auth_errmsg);
+	    gdm_error (_("Couldn't authenticate user \"%s\""), login);
+
+	    print_cant_auth_errbox ();
+
 	    g_free (login);
 	    g_free (passwd);
 	    g_free (ppasswd);
@@ -131,16 +182,10 @@ gdm_verify_user (GdmDisplay *d,
     if (ppasswd == NULL || (ppasswd[0] != '\0' &&
 			    strcmp (crypt (passwd, ppasswd), ppasswd) != 0)) {
 	    sleep (GdmRetryDelay);
-	    /* FIXME: Hmm, how are we sure that the login is username
-	     * and password.  That is the most common case but not
-	     * neccessairly true, this message needs to be changed
-	     * to allow for such cases */
-	    auth_errmsg = g_strdup_printf
-		    (_("\nIncorrect username or password.  "
-		       "Letters must be typed in the correct case.  "  
-		       "Please be sure the Caps Lock key is not enabled"));
-	    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX, auth_errmsg);
-	    g_free (auth_errmsg);
+	    gdm_error (_("Couldn't authenticate user \"%s\""), login);
+
+	    print_cant_auth_errbox ();
+
 	    g_free (login);
 	    g_free (passwd);
 	    g_free (ppasswd);
@@ -208,10 +253,13 @@ gdm_verify_user (GdmDisplay *d,
 
 gboolean
 gdm_verify_setup_user (GdmDisplay *d,
-		       const gchar *login, const gchar *display) 
+		       const gchar *login, const gchar *display,
+		       char **new_login) 
 {
 	struct passwd *pwent;
 
+	*new_login = NULL;
+
 	pwent = getpwnam (login);
 	if (pwent == NULL) {
 		gdm_error (_("Cannot get passwd structure for %s"), login);
