@@ -27,6 +27,10 @@
 #include <security/pam_appl.h>
 #include <pwd.h>
 
+#ifdef HAVE_DEFOPEN
+#include <deflt.h>
+#endif
+
 #include <vicious.h>
 
 #include "gdm.h"
@@ -48,6 +52,62 @@ static pam_handle_t *pamh = NULL;
 
 static GdmDisplay *cur_gdm_disp = NULL;
 
+/* this is a hack */
+static char *tmp_PAM_USER = NULL;
+
+/* this is another hack */
+static gboolean did_we_ask_for_password = FALSE;
+
+static char *selected_user = NULL;
+
+void
+gdm_verify_select_user (const char *user)
+{
+	g_free (selected_user);
+	if (ve_string_empty (user))
+		selected_user = NULL;
+	else
+		selected_user = g_strdup (user);
+}
+
+static const char *
+perhaps_translate_message (const char *msg)
+{
+	char *s;
+	const char *ret;
+	static GHashTable *hash = NULL;
+	if (hash == NULL) {
+		/* Here we come with some fairly standard messages so that
+		   we have as much as possible translated.  Should really be
+		   translated in pam I suppose */
+		hash = g_hash_table_new (g_str_hash, g_str_equal);
+		/* login: is whacked always translate to Username: */
+		g_hash_table_insert (hash, "login:", _("Username:"));
+		g_hash_table_insert (hash, "Username:", _("Username:"));
+		g_hash_table_insert (hash, "Password:", _("Password:"));
+		g_hash_table_insert (hash, "You are required to change your password immediately (password aged)", _("You are required to change your password immediately (password aged)"));
+		g_hash_table_insert (hash, "You are required to change your password immediately (root enforced)", _("You are required to change your password immediately (root enforced)"));
+		g_hash_table_insert (hash, "Your account has expired; please contact your system administrator", _("Your account has expired; please contact your system administrator"));
+		g_hash_table_insert (hash, "No password supplied", _("No password supplied"));
+		g_hash_table_insert (hash, "Password unchanged", _("Password unchanged"));
+		g_hash_table_insert (hash, "Can not get username", _("Can not get username"));
+		g_hash_table_insert (hash, "Retype new UNIX password:", _("Retype new UNIX password:"));
+		g_hash_table_insert (hash, "Enter new UNIX password:", _("Enter new UNIX password:"));
+		g_hash_table_insert (hash, "(current) UNIX password:", _("(current) UNIX password:"));
+		g_hash_table_insert (hash, "Error while changing NIS password.", _("Error while changing NIS password."));
+		g_hash_table_insert (hash, "You must choose a longer password", _("You must choose a longer password"));
+		g_hash_table_insert (hash, "Password has been already used. Choose another.", _("Password has been already used. Choose another."));
+		g_hash_table_insert (hash, "You must wait longer to change your password", _("You must wait longer to change your password"));
+		g_hash_table_insert (hash, "Sorry, passwords do not match", _("Sorry, passwords do not match"));
+	}
+	s = g_strstrip (g_strdup (msg));
+	ret = g_hash_table_lookup (hash, s);
+	g_free (s);
+	if (ret != NULL)
+		return ret;
+	else
+		return msg;
+}
 
 /* Internal PAM conversation function. Interfaces between the PAM
  * authentication system and the actual greeter program */
@@ -58,8 +118,10 @@ gdm_verify_pam_conv (int num_msg, const struct pam_message **msg,
 		     void *appdata_ptr)
 {
     int replies = 0;
+    int i;
     char *s;
     struct pam_response *reply = NULL;
+    const char *login;
     
     reply = malloc (sizeof (struct pam_response) * num_msg);
     
@@ -67,31 +129,83 @@ gdm_verify_pam_conv (int num_msg, const struct pam_message **msg,
 	return PAM_CONV_ERR;
 
     memset (reply, 0, sizeof (struct pam_response) * num_msg);
+
+    /* Here we set the login if it wasn't already set,
+     * this is kind of anal, but this way we guarantee that
+     * the greeter always is up to date on the login */
+    if (pam_get_item (pamh, PAM_USER, (const void **)&login) == PAM_SUCCESS ||
+	tmp_PAM_USER != NULL) {
+	    /* FIXME: this is a HACK HACK HACK, for some reason needed */
+	    if (tmp_PAM_USER != NULL)
+		    login = tmp_PAM_USER;
+	    gdm_slave_greeter_ctl_no_ret (GDM_SETLOGIN, login);
+    }
     
     for (replies = 0; replies < num_msg; replies++) {
+	gboolean islogin = FALSE;
+	const char *m = msg[replies]->msg;
+	m = perhaps_translate_message (m);
 	
 	switch (msg[replies]->msg_style) {
 	    
+	/* PAM requested textual input with echo on */
 	case PAM_PROMPT_ECHO_ON:
-	    /* PAM requested textual input with echo on */
-	    s = gdm_slave_greeter_ctl (GDM_PROMPT, _(msg[replies]->msg));
+ 	    if (strcmp(m, _("Username:")) == 0) {
+		    /* this is an evil hack, but really there is no way we'll
+		       know this is a username prompt.  However we SHOULD NOT
+		       rely on this working.  The pam modules can set their
+		       prompt to whatever they wish to */
+		    gdm_slave_greeter_ctl_no_ret
+			    (GDM_MSG, _("Please enter your username"));
+		    s = gdm_slave_greeter_ctl (GDM_PROMPT, m);
+		    /* this will clear the message */
+		    gdm_slave_greeter_ctl_no_ret (GDM_MSG, "");
+		    islogin = TRUE;
+	    } else {
+		    s = gdm_slave_greeter_ctl (GDM_PROMPT, m);
+	    }
 
 	    if (gdm_slave_greeter_check_interruption ()) {
 		    g_free (s);
+		    for (i = 0; i < replies; i++)
+			    if (reply[replies].resp != NULL)
+				    free (reply[replies].resp);
 		    free (reply);
 		    return PAM_CONV_ERR;
 	    }
 
+	    if (islogin) {
+		    /* note that we should NOT rely on this for anything
+		       that is REALLY needed.  The reason this is here
+		       is for the face browser to work correctly.  The
+		       best way would really be to ask all questions at
+		       once, but that would 1) need some protocol rethinking,
+		       2) the gdmgreeter program is really not equipped
+		       for this, would need theme format change */
+		    /* Second problem is that pam will not actually set
+		       PAM_USER itself after it asks for it.  This is
+		       pretty evil, evil, evil, evil, but somewhat necessary
+		       then. */
+		    /* FIXME: this is a HACK HACK HACK */
+		    g_free (tmp_PAM_USER);
+		    tmp_PAM_USER = g_strdup (s);
+	    }
+
 	    reply[replies].resp_retcode = PAM_SUCCESS;
 	    reply[replies].resp = strdup (ve_sure_string (s));
 	    g_free (s);
 	    break;
 	    
 	case PAM_PROMPT_ECHO_OFF:
+	    if (strcmp (m, _("Password:")) == 0)
+		    did_we_ask_for_password = TRUE;
 	    /* PAM requested textual input with echo off */
-	    s = gdm_slave_greeter_ctl (GDM_NOECHO, _(msg[replies]->msg));
+	    s = gdm_slave_greeter_ctl (GDM_NOECHO, m);
 	    if (gdm_slave_greeter_check_interruption ()) {
 		    g_free (s);
+		    for (i = 0; i < replies; i++)
+			    if (reply[replies].resp != NULL)
+				    free (reply[replies].resp);
 		    free (reply);
 		    return PAM_CONV_ERR;
 	    }
@@ -102,19 +216,22 @@ gdm_verify_pam_conv (int num_msg, const struct pam_message **msg,
 	    
 	case PAM_ERROR_MSG:
 	    /* PAM sent a message that should displayed to the user */
-	    gdm_slave_greeter_ctl (GDM_ERRDLG, _(msg[replies]->msg));
+	    gdm_slave_greeter_ctl (GDM_ERRDLG, m);
 	    reply[replies].resp_retcode = PAM_SUCCESS;
 	    reply[replies].resp = NULL;
 	    break;
 	case PAM_TEXT_INFO:
 	    /* PAM sent a message that should displayed to the user */
-	    gdm_slave_greeter_ctl (GDM_MSG, _(msg[replies]->msg));
+	    gdm_slave_greeter_ctl (GDM_MSG, m);
 	    reply[replies].resp_retcode = PAM_SUCCESS;
 	    reply[replies].resp = NULL;
 	    break;
 	    
 	default:
 	    /* PAM has been smoking serious crack */
+	    for (i = 0; i < replies; i++)
+		    if (reply[replies].resp != NULL)
+			    free (reply[replies].resp);
 	    free (reply);
 	    return PAM_CONV_ERR;
 	}
@@ -131,13 +248,17 @@ static struct pam_conv pamc = {
     NULL
 };
 
+/* Extra message to give on queries */
+static char *extra_standalone_message = NULL;
+
 static gint 
 gdm_verify_standalone_pam_conv (int num_msg, const struct pam_message **msg,
 				struct pam_response **resp,
 				void *appdata_ptr)
 {
 	int replies = 0;
-	char *s;
+	int i;
+	char *s, *text;
 	struct pam_response *reply = NULL;
 
 	reply = malloc (sizeof (struct pam_response) * num_msg);
@@ -148,13 +269,22 @@ gdm_verify_standalone_pam_conv (int num_msg, const struct pam_message **msg,
 	memset (reply, 0, sizeof (struct pam_response) * num_msg);
 
 	for (replies = 0; replies < num_msg; replies++) {
+		const char *m = msg[replies]->msg;
+		m = perhaps_translate_message (m);
 		switch (msg[replies]->msg_style) {
 
 		case PAM_PROMPT_ECHO_ON:
+			if (extra_standalone_message != NULL)
+				text = g_strdup_printf
+					("%s\n%s", extra_standalone_message,
+					 m);
+			else
+				text = g_strdup (m);
+
 			/* PAM requested textual input with echo on */
-			s = gdm_failsafe_question (cur_gdm_disp,
-						   _(msg[replies]->msg),
+			s = gdm_failsafe_question (cur_gdm_disp, text,
 						   TRUE /* echo */);
+			g_free (text);
 
 			reply[replies].resp_retcode = PAM_SUCCESS;
 			reply[replies].resp = strdup (ve_sure_string (s));
@@ -162,11 +292,19 @@ gdm_verify_standalone_pam_conv (int num_msg, const struct pam_message **msg,
 			break;
 
 		case PAM_PROMPT_ECHO_OFF:
+			if (extra_standalone_message != NULL)
+				text = g_strdup_printf
+					("%s\n%s", extra_standalone_message,
+					 m);
+			else
+				text = g_strdup (m);
+
 			/* PAM requested textual input with echo off */
-			s = gdm_failsafe_question (cur_gdm_disp,
-						   _(msg[replies]->msg),
+			s = gdm_failsafe_question (cur_gdm_disp, text,
 						   FALSE /* echo */);
 
+			g_free (text);
+
 			reply[replies].resp_retcode = PAM_SUCCESS;
 			reply[replies].resp = strdup (ve_sure_string (s));
 			g_free (s);
@@ -176,7 +314,7 @@ gdm_verify_standalone_pam_conv (int num_msg, const struct pam_message **msg,
 			/* PAM sent a message that should displayed to the user */
 			gdm_error_box (cur_gdm_disp,
 				       GTK_MESSAGE_ERROR,
-				       _(msg[replies]->msg));
+				       m);
 			reply[replies].resp_retcode = PAM_SUCCESS;
 			reply[replies].resp = NULL;
 			break;
@@ -185,13 +323,16 @@ gdm_verify_standalone_pam_conv (int num_msg, const struct pam_message **msg,
 			/* PAM sent a message that should displayed to the user */
 			gdm_error_box (cur_gdm_disp,
 				       GTK_MESSAGE_INFO,
-				       _(msg[replies]->msg));
+				       m);
 			reply[replies].resp_retcode = PAM_SUCCESS;
 			reply[replies].resp = NULL;
 			break;
 
 		default:
 			/* PAM has been smoking serious crack */
+			for (i = 0; i < replies; i++)
+				if (reply[replies].resp != NULL)
+					free (reply[replies].resp);
 			free (reply);
 			return PAM_CONV_ERR;
 		}
@@ -216,10 +357,8 @@ create_pamh (GdmDisplay *d,
 	     const char *display,
 	     int *pamerr)
 {
-	if (login == NULL ||
-	    display == NULL) {
-		gdm_error (_("Cannot setup pam handle with null login "
-			     "and/or display"));
+	if (display == NULL) {
+		gdm_error (_("Cannot setup pam handle with null display"));
 		return FALSE;
 	}
 
@@ -232,7 +371,8 @@ create_pamh (GdmDisplay *d,
 	/* Initialize a PAM session for the user */
 	if ((*pamerr = pam_start (service, login, conv, &pamh)) != PAM_SUCCESS) {
 		if (gdm_slave_should_complain ())
-			gdm_error (_("Can't find /etc/pam.d/%s!"), service);
+			gdm_error (_("Unable to establish service %s: %s\n"),
+				   service, pam_strerror (NULL, *pamerr));
 		return FALSE;
 	}
 
@@ -243,20 +383,27 @@ create_pamh (GdmDisplay *d,
 		return FALSE;
 	}
 
+#if 0
+	/* Apparently we should not set RUSER, docs are unclear but the sun
+           guys are saying don't do it */
 	/* gdm is requesting the login */
 	if ((*pamerr = pam_set_item (pamh, PAM_RUSER, GdmUser)) != PAM_SUCCESS) {
 		if (gdm_slave_should_complain ())
 			gdm_error (_("Can't set PAM_RUSER=%s"), GdmUser);
 		return FALSE;
 	}
+#endif
 
-	/* From the host of the display */
-	if ((*pamerr = pam_set_item (pamh, PAM_RHOST,
-				     d->console ? "localhost" : d->hostname)) != PAM_SUCCESS) {
-		if (gdm_slave_should_complain ())
-			gdm_error (_("Can't set PAM_RHOST=%s"),
-				   d->console ? "localhost" : d->hostname);
-		return FALSE;
+	if ( ! d->console) {
+		/* Only set RHOST if host is remote */
+		/* From the host of the display */
+		if ((*pamerr = pam_set_item (pamh, PAM_RHOST,
+					     d->hostname)) != PAM_SUCCESS) {
+			if (gdm_slave_should_complain ())
+				gdm_error (_("Can't set PAM_RHOST=%s"),
+					   d->hostname);
+			return FALSE;
+		}
 	}
 
 	return TRUE;
@@ -282,12 +429,12 @@ gdm_verify_user (GdmDisplay *d,
 		 gboolean local) 
 {
     gint pamerr = 0;
-    gchar *login;
+    char *login = NULL;
     struct passwd *pwent;
     gboolean error_msg_given = FALSE;
     gboolean credentials_set = FALSE;
     gboolean started_timer = FALSE;
-    gchar *auth_errmsg;
+    int null_tok = 0;
 
     /* start the timer for timed logins */
     if ( ! ve_string_empty (GdmTimedLogin) &&
@@ -296,50 +443,107 @@ gdm_verify_user (GdmDisplay *d,
 	    started_timer = TRUE;
     }
 
-    if (username == NULL) {
-	    /* Ask gdmgreeter for the user's login. Just for good measure */
-	    gdm_slave_greeter_ctl_no_ret (GDM_MSG, _("Please enter your username"));
-	    login = gdm_slave_greeter_ctl (GDM_LOGIN, _("Username:"));
-	    if (login == NULL ||
-		gdm_slave_greeter_check_interruption ()) {
-		    if (started_timer)
-			    gdm_slave_greeter_ctl_no_ret (GDM_STOPTIMER, "");
-		    g_free (login);
-		    return NULL;
-	    }
-    } else {
+    if (username != NULL) {
 	    login = g_strdup (username);
+	    gdm_slave_greeter_ctl_no_ret (GDM_SETLOGIN, login);
     }
 
     cur_gdm_disp = d;
 
     /* Initialize a PAM session for the user */
-    if ( ! create_pamh (d, "gdm", login, &pamc, display, &pamerr)) {
+    if ( ! create_pamh (d, "gdm", username, &pamc, display, &pamerr)) {
 	    if (started_timer)
 		    gdm_slave_greeter_ctl_no_ret (GDM_STOPTIMER, "");
 	    goto pamerr;
     }
+
+    pam_set_item (pamh, PAM_USER_PROMPT, _("Username:"));
+
+    /* A Solaris thing: */
+#ifdef HAVE_DEFOPEN
+    if (defopen(DEFLT"/login") == 0) {
+	    char *passreq;
+	    int def_flgs;
+
+	    def_flgs = defcntl(DC_GETFLAGS, 0);
+	    TURNOFF(def_flgs, DC_CASE);
+	    (void) defcntl (DC_SETFLAGS, def_flgs);	 /* ignore case */
+
+	    if (((passreq = defread("PASSREQ=")) != NULL) &&
+		g_ascii_strcasecmp (passreq, "YES") == 0)
+		    null_tok |= PAM_DISALLOW_NULL_AUTHTOK;
+
+	    (void) defopen(NULL); /* close defaults file  */
+    }
+#endif
 	    
 #ifdef PAM_FAIL_DELAY
     pam_fail_delay (pamh, GdmRetryDelay * 1000);
 #endif /* PAM_FAIL_DELAY */
 
+authenticate_again:
+    /* hack */
+    g_free (tmp_PAM_USER);
+    tmp_PAM_USER = NULL;
+
+    did_we_ask_for_password = FALSE;
+
+    gdm_verify_select_user (NULL);
     /* Start authentication session */
-    if ((pamerr = pam_authenticate (pamh, 0)) != PAM_SUCCESS) {
-#ifndef PAM_FAIL_DELAY
-	    sleep (GdmRetryDelay);
-#endif /* PAM_FAIL_DELAY */
+    if ((pamerr = pam_authenticate (pamh, null_tok)) != PAM_SUCCESS) {
+	    if ( ! ve_string_empty (selected_user)) {
+		    /* FIXME: what about errors */
+		    pam_set_item (pamh, PAM_USER, selected_user);
+		    /* Note that the GDM_SETUSER will be sent in the
+		       authenticate conversation.  This is a more robust
+		       solution. */
+		    goto authenticate_again;
+	    }
 	    if (started_timer)
 		    gdm_slave_greeter_ctl_no_ret (GDM_STOPTIMER, "");
-	    if (gdm_slave_should_complain ())
+	    if (gdm_slave_should_complain ()) {
+#ifndef PAM_FAIL_DELAY
+		    sleep (GdmRetryDelay);
+#endif /* PAM_FAIL_DELAY */
 		    gdm_error (_("Couldn't authenticate user"));
+	    }
 	    goto pamerr;
     }
 
     /* stop the timer for timed logins */
     if (started_timer)
 	    gdm_slave_greeter_ctl_no_ret (GDM_STOPTIMER, "");
+
+    g_free (login);
+    login = NULL;
     
+    if ((pamerr = pam_get_item (pamh, PAM_USER, (const void **)&login))
+	!= PAM_SUCCESS) {
+	    login = NULL;
+	    /* is not really an auth problem, but it will
+	       pretty much look as such, it shouldn't really
+	       happen */
+	    if (gdm_slave_should_complain ())
+		    gdm_error (_("Couldn't authenticate user"));
+	    goto pamerr;
+    }
+    login = g_strdup (login);
+    /* kind of anal, the greeter likely already knows, but it could have
+       been changed */
+    gdm_slave_greeter_ctl_no_ret (GDM_SETLOGIN, login);
+
+    if ( ! gdm_slave_check_user_wants_to_log_in (login)) {
+	    /* cleanup stuff */
+	    gdm_slave_greeter_ctl_no_ret (GDM_SETLOGIN, "");
+	    g_free (login);
+	    login = NULL;
+	    gdm_slave_greeter_ctl_no_ret (GDM_RESETOK, "");
+
+	    gdm_verify_cleanup (d);
+	    /* kind of a hack */
+	    return gdm_verify_user (d, username, display, local);
+    }
+
     pwent = getpwnam (login);
     if ( ( ! GdmAllowRoot ||
 	  ( ! GdmAllowRemoteRoot && ! local) ) &&
@@ -358,7 +562,7 @@ gdm_verify_user (GdmDisplay *d,
     }
 
     /* Check if the user's account is healthy. */
-    pamerr = pam_acct_mgmt (pamh, 0);
+    pamerr = pam_acct_mgmt (pamh, null_tok);
     switch (pamerr) {
     case PAM_SUCCESS :
 	break;
@@ -436,15 +640,39 @@ gdm_verify_user (GdmDisplay *d,
 	    /* I'm not sure yet if I should display this message for any other issues - heeten */
 	    if (pamerr == PAM_AUTH_ERR ||
 		pamerr == PAM_USER_UNKNOWN) {
-		    /* FIXME: Hmm, how are we sure that the login is username
-		     * and password.  That is the most common case but not
-		     * neccessairly true, this message needs to be changed
-		     * to allow for such cases */
-		    auth_errmsg = 
-			    _("\nIncorrect username or password. "
-			      "Letters must be typed in the correct case. "  
-			      "Please make sure the Caps Lock key is not enabled.");
-		    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX, auth_errmsg);
+		    gboolean is_capslock = FALSE;
+		    const char *basemsg;
+		    char *msg;
+		    char *ret;
+			    
+		    ret = gdm_slave_greeter_ctl (GDM_QUERY_CAPSLOCK, "");
+		    if ( ! ve_string_empty (ret))
+			    is_capslock = TRUE;
+		    g_free (ret);
+
+		    /* Only give this message if we actually asked for
+		       password, otherwise it would be silly to say that
+		       the password may have been wrong */
+		    if (did_we_ask_for_password) {
+			    basemsg = _("\nIncorrect username or password.  "
+					"Letters must be typed in the correct "
+					"case.");
+		    } else {
+			    basemsg = _("\nAuthentication failed.  "
+					"Letters must be typed in the correct "
+					"case.");
+		    }
+		    if (is_capslock) {
+			    msg = g_strconcat (basemsg, "  ",
+					       _("Please make sure the "
+						 "Caps Lock key is not "
+						 "enabled."),
+					       NULL);
+		    } else {
+			    msg = g_strdup (basemsg);
+		    }
+		    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX, msg);
+		    g_free (msg);
 	    } else {
 		    gdm_slave_greeter_ctl_no_ret (GDM_ERRDLG, _("Authentication failed"));
 	    }
@@ -479,24 +707,53 @@ gdm_verify_user (GdmDisplay *d,
  */
 
 gboolean
-gdm_verify_setup_user (GdmDisplay *d, const gchar *login, const gchar *display) 
+gdm_verify_setup_user (GdmDisplay *d, const gchar *login, const gchar *display,
+		       char **new_login) 
 {
     gint pamerr = 0;
     struct passwd *pwent;
+    const char *after_login;
+    int null_tok = 0;
+    
+    *new_login = NULL;
 
     if (login == NULL)
 	    return FALSE;
 
     cur_gdm_disp = d;
 
+    g_free (extra_standalone_message);
+    extra_standalone_message = g_strdup_printf ("%s (%s)",
+						_("Automatic login"),
+						login);
+
     /* Initialize a PAM session for the user */
     if ( ! create_pamh (d, "gdm-autologin", login, &standalone_pamc,
 			display, &pamerr)) {
 	    goto setup_pamerr;
     }
 
+    /* A Solaris thing: */
+#ifdef HAVE_DEFOPEN
+    if (defopen(DEFLT"/login") == 0) {
+	    char *passreq;
+	    int def_flgs;
+
+	    def_flgs = defcntl(DC_GETFLAGS, 0);
+	    TURNOFF(def_flgs, DC_CASE);
+	    (void) defcntl (DC_SETFLAGS, def_flgs);	 /* ignore case */
+
+	    if (((passreq = defread("PASSREQ=")) != NULL) &&
+		g_ascii_strcasecmp (passreq, "YES") == 0)
+		    null_tok |= PAM_DISALLOW_NULL_AUTHTOK;
+
+	    (void) defopen(NULL); /* close defaults file  */
+    }
+#endif
+
     /* Start authentication session */
-    if ((pamerr = pam_authenticate (pamh, 0)) != PAM_SUCCESS) {
+    did_we_ask_for_password = FALSE;
+    if ((pamerr = pam_authenticate (pamh, null_tok)) != PAM_SUCCESS) {
 	    if (gdm_slave_should_complain ()) {
 		    gdm_error (_("Couldn't authenticate user"));
 		    gdm_error_box (cur_gdm_disp,
@@ -506,8 +763,25 @@ gdm_verify_setup_user (GdmDisplay *d, const gchar *login, const gchar *display)
 	    goto setup_pamerr;
     }
 
+    if ((pamerr = pam_get_item (pamh, PAM_USER, (const void **)&after_login))
+	!= PAM_SUCCESS) {
+	    /* is not really an auth problem, but it will
+	       pretty much look as such, it shouldn't really
+	       happen */
+	    gdm_error (_("Couldn't authenticate user"));
+	    gdm_error_box (cur_gdm_disp,
+			   GTK_MESSAGE_ERROR,
+			   _("Authentication failed"));
+	    goto setup_pamerr;
+    }
+
+    if (after_login != NULL /* should never be */ &&
+	strcmp (after_login, login) != 0) {
+	    *new_login = g_strdup (after_login);
+    }
+
     /* Check if the user's account is healthy. */
-    pamerr = pam_acct_mgmt (pamh, 0);
+    pamerr = pam_acct_mgmt (pamh, null_tok);
     switch (pamerr) {
     case PAM_SUCCESS :
 	break;
@@ -557,14 +831,6 @@ gdm_verify_setup_user (GdmDisplay *d, const gchar *login, const gchar *display)
 	    goto setup_pamerr;
     }
 
-    /* Register the session */
-    pamerr = pam_open_session (pamh, 0);
-    if (pamerr != PAM_SUCCESS) {
-	    if (gdm_slave_should_complain ())
-		    gdm_error (_("Couldn't open session for %s"), login);
-	    goto setup_pamerr;
-    }
-
     /* Set credentials */
     pamerr = pam_setcred (pamh, PAM_ESTABLISH_CRED);
     if (pamerr != PAM_SUCCESS) {
@@ -573,11 +839,25 @@ gdm_verify_setup_user (GdmDisplay *d, const gchar *login, const gchar *display)
 	    goto setup_pamerr;
     }
 
+    /* Register the session */
+    pamerr = pam_open_session (pamh, 0);
+    if (pamerr != PAM_SUCCESS) {
+	    /* Throw away the credentials */
+	    pam_setcred (pamh, PAM_DELETE_CRED);
+
+	    if (gdm_slave_should_complain ())
+		    gdm_error (_("Couldn't open session for %s"), login);
+	    goto setup_pamerr;
+    }
+
     /* Workaround to avoid gdm messages being logged as PAM_pwdb */
     closelog ();
     openlog ("gdm", LOG_PID, LOG_DAEMON);
 
     cur_gdm_disp = NULL;
+
+    g_free (extra_standalone_message);
+    extra_standalone_message = NULL;
     
     return TRUE;
     
@@ -593,6 +873,9 @@ gdm_verify_setup_user (GdmDisplay *d, const gchar *login, const gchar *display)
 
     cur_gdm_disp = NULL;
 
+    g_free (extra_standalone_message);
+    extra_standalone_message = NULL;
+
     return FALSE;
 }
 
