@@ -21,6 +21,7 @@
 #include "greeter_geometry.h"
 #include "greeter_item_clock.h"
 #include "greeter_item_pam.h"
+#include "greeter_item_ulist.h"
 #include "greeter_item_capslock.h"
 #include "greeter_item_timed.h"
 #include "greeter_events.h"
@@ -29,18 +30,24 @@
 #include "greeter_system.h"
 
 gboolean DOING_GDM_DEVELOPMENT = FALSE;
+static char *greeter_Welcome_key = GDM_KEY_WELCOME;
 
 GtkWidget *window;
 GtkWidget *canvas;
 
+gboolean GDM_IS_LOCAL = FALSE;
 char *GdmGraphicalTheme = NULL;
 char *GdmGraphicalThemeDir = NULL;
 int GdmXineramaScreen = 0;
+#if 0
+/* FIXME: maybe just whack this */
 gboolean GdmShowGnomeChooserSession = FALSE;
+#endif
 gboolean GdmShowGnomeFailsafeSession = FALSE;
 gboolean GdmShowXtermFailsafeSession = FALSE;
 gboolean GdmShowLastSession = FALSE;
 gchar *GdmSessionDir = NULL;
+gchar *GdmDefaultSession = NULL;
 gchar *GdmLocaleFile = NULL;
 gchar *GdmHalt = NULL;
 gchar *GdmReboot = NULL;
@@ -48,19 +55,33 @@ gchar *GdmSuspend = NULL;
 gchar *GdmConfigurator = NULL;
 gboolean GdmSystemMenu = TRUE;
 gboolean GdmConfigAvailable = TRUE;
+gboolean GdmChooserButton = TRUE;
 gboolean GdmTimedLoginEnable;
 gboolean GdmUse24Clock;
+gchar *GdmGlobalFaceDir;
+gchar *GdmDefaultFace;
+gint  GdmIconMaxHeight;
+gint  GdmIconMaxWidth;
 gchar *GdmTimedLogin;
 gchar *GdmGtkRC;
 gint GdmTimedLoginDelay;
+gchar *GdmExclude;
+int GdmMinimalUID;
+gboolean GdmAllowRoot;
+gboolean GdmAllowRemoteRoot;
+gchar *GdmWelcome;
 
 gboolean GdmUseCirclesInEntry = FALSE;
 
 static gboolean used_defaults = FALSE;
 gint greeter_current_delay = 0;
 
+/* FIXME: hack */
+GreeterItemInfo *welcome_string_info = NULL;
+
 extern gboolean session_dir_whacked_out;
 
+gboolean greeter_probably_login_prompt = FALSE;
 
 static void 
 greeter_parse_config (void)
@@ -69,10 +90,17 @@ greeter_parse_config (void)
 
     if (!g_file_test (GDM_CONFIG_FILE, G_FILE_TEST_EXISTS))
       {
-	syslog (LOG_ERR, _("greeter_parse_config: No configuration file: %s. Using defaults."), GDM_CONFIG_FILE);
+	syslog (LOG_ERR, _("%s: No configuration file: %s. Using defaults."), 
+		"greeter_parse_config", GDM_CONFIG_FILE);
 	used_defaults = TRUE;
       }
 
+    if (ve_string_empty (g_getenv ("GDM_IS_LOCAL"))) {
+	    greeter_Welcome_key = GDM_KEY_REMOTEWELCOME;
+    } else {
+	    greeter_Welcome_key = GDM_KEY_WELCOME;
+    }
+
     config = ve_config_get (GDM_CONFIG_FILE);
 
     GdmGraphicalTheme = ve_config_get_string (config, GDM_KEY_GRAPHICAL_THEME);
@@ -95,19 +123,38 @@ greeter_parse_config (void)
 
     GdmShowXtermFailsafeSession = ve_config_get_bool (config, GDM_KEY_SHOW_XTERM_FAILSAFE);
     GdmShowGnomeFailsafeSession = ve_config_get_bool (config, GDM_KEY_SHOW_GNOME_FAILSAFE);
+#if 0
+    /* FIXME: maybe just whack this */
     GdmShowGnomeChooserSession = ve_config_get_bool (config, GDM_KEY_SHOW_GNOME_CHOOSER);
+#endif
     GdmShowLastSession = ve_config_get_bool (config, GDM_KEY_SHOW_LAST_SESSION);
     GdmSessionDir = ve_config_get_string (config, GDM_KEY_SESSDIR);
+    GdmDefaultSession = ve_config_get_string (config, GDM_KEY_DEFAULTSESSION);
     GdmLocaleFile = ve_config_get_string (config, GDM_KEY_LOCFILE);
     GdmSystemMenu = ve_config_get_bool (config, GDM_KEY_SYSMENU);
     GdmConfigAvailable = ve_config_get_bool (config, GDM_KEY_CONFIG_AVAILABLE);
+    GdmChooserButton = ve_config_get_bool (config, GDM_KEY_CHOOSER_BUTTON);
     GdmHalt = ve_config_get_string (config, GDM_KEY_HALT);
     GdmReboot = ve_config_get_string (config, GDM_KEY_REBOOT);
     GdmSuspend = ve_config_get_string (config, GDM_KEY_SUSPEND);
     GdmConfigurator = ve_config_get_string (config, GDM_KEY_CONFIGURATOR);
     GdmGtkRC = ve_config_get_string (config, GDM_KEY_GTKRC);
 
+    GdmWelcome = ve_config_get_translated_string (config, greeter_Welcome_key);
+    /* A hack! */
+    if (strcmp (ve_sure_string (GdmWelcome), "Welcome") == 0) {
+	    g_free (GdmWelcome);
+	    GdmWelcome = g_strdup (_("Welcome"));
+    } else if (strcmp (ve_sure_string (GdmWelcome), "Welcome to %n") == 0) {
+	    g_free (GdmWelcome);
+	    GdmWelcome = g_strdup (_("Welcome to %n"));
+    }
+
     GdmTimedLoginEnable = ve_config_get_bool (config, GDM_KEY_TIMED_LOGIN_ENABLE);
+    GdmExclude = ve_config_get_string (config, GDM_KEY_EXCLUDE);
+    GdmMinimalUID = ve_config_get_int (config, GDM_KEY_MINIMALUID);
+    GdmAllowRoot = ve_config_get_bool (config, GDM_KEY_ALLOWROOT);
+    GdmAllowRemoteRoot = ve_config_get_bool (config, GDM_KEY_ALLOWREMOTEROOT);
     GdmTimedLoginDelay = ve_config_get_int (config, GDM_KEY_TIMED_LOGIN_DELAY);
 
     /* Note: TimedLogin here is not gotten out of the config
@@ -125,7 +172,7 @@ greeter_parse_config (void)
 	if (GdmTimedLoginDelay < 5)
 	  {
 	    syslog (LOG_WARNING,
-		    _("TimedLoginDelay was less then 5.  I'll just use 5."));
+		    _("TimedLoginDelay was less than 5.  I'll just use 5."));
 	    GdmTimedLoginDelay = 5;
 	  }
       }
@@ -137,8 +184,20 @@ greeter_parse_config (void)
 
     GdmUse24Clock = ve_config_get_bool (config, GDM_KEY_USE_24_CLOCK);
 
+    GdmIconMaxWidth = ve_config_get_int (config, GDM_KEY_ICONWIDTH);
+    GdmIconMaxHeight = ve_config_get_int (config, GDM_KEY_ICONHEIGHT);
+    GdmGlobalFaceDir = ve_config_get_string (config, GDM_KEY_FACEDIR);
+    GdmDefaultFace = ve_config_get_string (config, GDM_KEY_FACE);
+
     if (GdmXineramaScreen < 0)
       GdmXineramaScreen = 0;
+    if (GdmIconMaxWidth < 0) GdmIconMaxWidth = 128;
+    if (GdmIconMaxHeight < 0) GdmIconMaxHeight = 128;
+    if (ve_string_empty (g_getenv ("GDM_IS_LOCAL"))) {
+	    GDM_IS_LOCAL = FALSE;
+    } else {
+	    GDM_IS_LOCAL = TRUE;
+    }
 }
 
 static void
@@ -191,24 +250,24 @@ greeter_ctrl_handler (GIOChannel *source,
 	buf[len-1] = '\0';
 	
 	greeter_item_pam_set_user (buf);
+	if (ve_string_empty (buf))
+	  greeter_item_ulist_enable ();
+	else
+	  greeter_item_ulist_disable ();
 	printf ("%c\n", STX);
 	fflush (stdout);
 	break;
-    case GDM_LOGIN:
-        g_io_channel_read_chars (source, buf, PIPE_SIZE-1, &len, NULL);
-	buf[len-1] = '\0';
-
-	tmp = ve_locale_to_utf8 (buf);
-	greeter_item_pam_prompt (tmp, 32, TRUE, TRUE);
-	g_free (tmp);
-	break;
 
     case GDM_PROMPT:
         g_io_channel_read_chars (source, buf, PIPE_SIZE-1, &len, NULL);
 	buf[len-1] = '\0';
 
 	tmp = ve_locale_to_utf8 (buf);
-	greeter_item_pam_prompt (tmp, 128, TRUE, FALSE);
+	if (tmp != NULL && strcmp (tmp, _("Username:")) == 0)
+		greeter_probably_login_prompt = TRUE;
+	else
+		greeter_probably_login_prompt = FALSE;
+	greeter_item_pam_prompt (tmp, 128, TRUE);
 	g_free (tmp);
 	break;
 
@@ -217,7 +276,7 @@ greeter_ctrl_handler (GIOChannel *source,
 	buf[len-1] = '\0';
 
 	tmp = ve_locale_to_utf8 (buf);
-	greeter_item_pam_prompt (tmp, 128, FALSE, FALSE);
+	greeter_item_pam_prompt (tmp, 128, FALSE);
 	g_free (tmp);
 	break;
 
@@ -259,6 +318,7 @@ greeter_ctrl_handler (GIOChannel *source,
 				      "%s",
 				      tmp);
 	g_free (tmp);
+	gtk_dialog_set_has_separator (GTK_DIALOG (dlg), FALSE);
 
 	gdm_wm_center_window (GTK_WINDOW (dlg));
 
@@ -353,6 +413,8 @@ greeter_ctrl_handler (GIOChannel *source,
 	_exit (EXIT_SUCCESS);
 	break;
 
+#if 0
+	/* FIXME: Maybe whack this */
     case GDM_GNOMESESS:
 	{
 		char *sess;
@@ -390,6 +452,7 @@ greeter_ctrl_handler (GIOChannel *source,
 	fflush (stdout);
 
 	break;
+#endif
 
     case GDM_STARTTIMER:
 	g_io_channel_read_chars (source, buf, PIPE_SIZE-1, &len, NULL); /* Empty */
@@ -482,6 +545,17 @@ greeter_ctrl_handler (GIOChannel *source,
 	gdk_flush ();
 
 	_exit (EXIT_SUCCESS);
+
+    case GDM_QUERY_CAPSLOCK:
+        g_io_channel_read_chars (source, buf, PIPE_SIZE-1, &len, NULL); /* Empty */
+
+	if (greeter_is_capslock_on ())
+	    printf ("%cY\n", STX);
+	else
+	    printf ("%c\n", STX);
+	fflush (stdout);
+
+	break;
 	
     default:
 	break;
@@ -506,9 +580,12 @@ key_press_event (GtkWidget *widget, GdkEventKey *key, gpointer data)
 static void
 greeter_setup_items (void)
 {
-
   greeter_item_clock_setup ();
   greeter_item_pam_setup ();
+
+  /* This will query the daemon for pictures through stdin/stdout! */
+  greeter_item_ulist_setup ();
+
   greeter_item_capslock_setup (window);
   greeter_item_timed_setup ();
   greeter_item_register_action_callback ("language_button",
@@ -558,6 +635,7 @@ verify_gdm_version (void)
 					 "You have probably just upgraded gdm.\n"
 					 "Please restart the gdm daemon or reboot the computer."),
 				       VERSION);
+      gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
     
       gtk_widget_show_all (dialog);
       gdm_wm_center_window (GTK_WINDOW (dialog));
@@ -587,6 +665,7 @@ verify_gdm_version (void)
 					 "You have probably just upgraded gdm.\n"
 					 "Please restart the gdm daemon or reboot the computer."),
 				       VERSION);
+      gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
       gtk_dialog_add_buttons (GTK_DIALOG (dialog),
 			      _("Reboot"),
 			      RESPONSE_REBOOT,
@@ -630,6 +709,7 @@ verify_gdm_version (void)
 					 "You have probably just upgraded gdm.\n"
 					 "Please restart the gdm daemon or reboot the computer."),
 				       VERSION, gdm_version);
+      gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
       gtk_dialog_add_buttons (GTK_DIALOG (dialog),
 			      _("Restart"),
 			      RESPONSE_RESTART,
@@ -678,6 +758,7 @@ greeter_message (const gchar *msg)
 				GTK_BUTTONS_CLOSE,
 				"%s",
 				msg);
+  gtk_dialog_set_has_separator (GTK_DIALOG (req), FALSE);
   g_signal_connect (G_OBJECT (req), "destroy",
 		    G_CALLBACK (gtk_widget_destroyed),
 		    &req);
@@ -708,6 +789,7 @@ greeter_query (const gchar *msg)
 				      GTK_BUTTONS_YES_NO,
 				      "%s",
 				      msg);
+	gtk_dialog_set_has_separator (GTK_DIALOG (req), FALSE);
 	gtk_label_set_use_markup
 		(GTK_LABEL (GTK_MESSAGE_DIALOG (req)->label), TRUE);
 
@@ -799,6 +881,7 @@ greeter_reread_config (int sig, gpointer data)
 {
 	VeConfig *config = ve_config_get (GDM_CONFIG_FILE);
 	char *theme, *theme_dir;
+	char *str;
 
 	theme = ve_config_get_string (config, GDM_KEY_GRAPHICAL_THEME);
 	if (ve_string_empty (theme)) {
@@ -832,9 +915,12 @@ greeter_reread_config (int sig, gpointer data)
 	     ! bool_same (config,
 			  GdmShowGnomeFailsafeSession,
 			  GDM_KEY_SHOW_GNOME_FAILSAFE) ||
+#if 0
+	     /* FIXME: maybe just whack this */
 	     ! bool_same (config,
 			  GdmShowGnomeChooserSession,
 			  GDM_KEY_SHOW_GNOME_CHOOSER) ||
+#endif
 	     ! string_same (config,
 			    GdmSessionDir,
 			    GDM_KEY_SESSDIR) ||
@@ -856,6 +942,12 @@ greeter_reread_config (int sig, gpointer data)
 	     ! string_same (config,
 			    GdmConfigurator,
 			    GDM_KEY_CONFIGURATOR) ||
+	     ! bool_same (config,
+			  GdmConfigAvailable,
+			  GDM_KEY_CONFIG_AVAILABLE) ||
+	     ! bool_same (config,
+			  GdmChooserButton,
+			  GDM_KEY_CHOOSER_BUTTON) ||
 	     ! bool_same (config,
 			  GdmTimedLoginEnable,
 			  GDM_KEY_TIMED_LOGIN_ENABLE) ||
@@ -881,6 +973,27 @@ greeter_reread_config (int sig, gpointer data)
 		greeter_item_clock_update ();
 	}
 
+	str = ve_config_get_translated_string (config, greeter_Welcome_key);
+	/* A hack */
+	if (strcmp (ve_sure_string (str), "Welcome") == 0) {
+		g_free (str);
+		str = g_strdup (_("Welcome"));
+	} else if (strcmp (ve_sure_string (str), "Welcome to %n") == 0) {
+		g_free (str);
+		str = g_strdup (_("Welcome to %n"));
+	}
+	if (strcmp (ve_sure_string (str), ve_sure_string (GdmWelcome)) != 0) {
+		g_free (GdmWelcome);
+		GdmWelcome = str;
+		if (welcome_string_info != NULL) {
+			g_free (welcome_string_info->orig_text);
+			welcome_string_info->orig_text = g_strdup (str);
+			greeter_item_update_text (welcome_string_info);
+		}
+	} else {
+		g_free (str);
+	}
+
 	g_free (theme);
 	g_free (theme_dir);
 
@@ -1141,6 +1254,7 @@ main (int argc, char *argv[])
 				         _("There was an error loading the "
 					   "theme %s"),
 					 g_getenv ("GDM_THEME"));
+	gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
     
         gtk_widget_show_all (dialog);
         gdm_wm_center_window (GTK_WINDOW (dialog));
@@ -1180,6 +1294,7 @@ main (int argc, char *argv[])
 					 "is corrupt.  It does not contain "
 					 "definition for the username/password "
 					 "entry element."));
+      gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
 
       gtk_widget_show_all (dialog);
       gdm_wm_center_window (GTK_WINDOW (dialog));
@@ -1214,6 +1329,7 @@ main (int argc, char *argv[])
 					 "also could not have been loaded, "
 					 "I will attempt to start the "
 					 "standard greeter"));
+      gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
     
       gtk_widget_show_all (dialog);
       gdm_wm_center_window (GTK_WINDOW (dialog));
@@ -1234,6 +1350,7 @@ main (int argc, char *argv[])
 					 "and you may have to login another "
 					 "way and fix the installation of "
 					 "gdm"));
+      gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
     
       gtk_widget_show_all (dialog);
       gdm_wm_center_window (GTK_WINDOW (dialog));
@@ -1279,6 +1396,7 @@ main (int argc, char *argv[])
 				       _("Your session directory is missing or empty!\n\n"
 					 "There are two available sessions you can use, but\n"
 					 "you should log in and correct the gdm configuration."));
+      gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
       gtk_widget_show_all (dialog);
       gdm_wm_center_window (GTK_WINDOW (dialog));
 
@@ -1303,6 +1421,7 @@ main (int argc, char *argv[])
 				       _("The configuration file contains an invalid command\n"
 					 "line for the login dialog, and thus I ran the\n"
 					 "default command.  Please fix your configuration."));
+      gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
       gtk_widget_show_all (dialog);
       gdm_wm_center_window (GTK_WINDOW (dialog));
 
@@ -1329,6 +1448,7 @@ main (int argc, char *argv[])
 					 "defaults to run this session.  You should log in\n"
 					 "and create a configuration file with the GDM\n"
 					 "configuration program."));
+      gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
       gtk_widget_show_all (dialog);
       gdm_wm_center_window (GTK_WINDOW (dialog));
 
