@@ -7,9 +7,11 @@
 #include <librsvg/rsvg.h>
 #include "vicious.h"
 
+#include "gdm.h"
 #include "greeter.h"
 #include "greeter_item.h"
 #include "greeter_events.h"
+#include "greeter_system.h"
 #include "greeter_canvas_item.h"
 #include "greeter_configuration.h"
 
@@ -119,29 +121,21 @@ make_menubar (void)
 	menu = gtk_menu_new();
 	gtk_menu_item_set_submenu (GTK_MENU_ITEM (w), menu);
 
-	w = gtk_menu_item_new_with_mnemonic (_("_Language"));
+	w = gtk_menu_item_new_with_mnemonic (_("Select _Language..."));
 	gtk_menu_shell_append (GTK_MENU_SHELL (menu), w);
 	gtk_widget_show (GTK_WIDGET (w));
 	g_signal_connect (G_OBJECT (w), "activate",
 			  G_CALLBACK (activate_button),
 			  "language_button");
 
-	w = gtk_menu_item_new_with_mnemonic (_("_Session"));
+	w = gtk_menu_item_new_with_mnemonic (_("Select _Session..."));
 	gtk_menu_shell_append (GTK_MENU_SHELL (menu), w);
 	gtk_widget_show (GTK_WIDGET (w));
 	g_signal_connect (G_OBJECT (w), "activate",
 			  G_CALLBACK (activate_button),
 			  "session_button");
 
-	if (GdmSystemMenu &&
-	    !  ve_string_empty (g_getenv ("GDM_IS_LOCAL"))) {
-		w = gtk_menu_item_new_with_mnemonic (_("S_ystem"));
-		gtk_menu_shell_append (GTK_MENU_SHELL (menu), w);
-		gtk_widget_show (GTK_WIDGET (w));
-		g_signal_connect (G_OBJECT (w), "activate",
-				  G_CALLBACK (activate_button),
-				  "system_button");
-	}
+	greeter_system_append_system_menu (menu);
 
 	/* Add a quit/disconnect item when in xdmcp mode or flexi mode */
 	/* Do note that the order is important, we always want "Quit" for
@@ -155,6 +149,12 @@ make_menubar (void)
 		w = NULL;
 	}
 	if (w != NULL) {
+		GtkWidget *sep;
+		/* add separator before the quit */
+		sep = gtk_separator_menu_item_new ();
+		gtk_menu_shell_append (GTK_MENU_SHELL (menu), sep);
+		gtk_widget_show (GTK_WIDGET (sep));
+
 		gtk_menu_shell_append (GTK_MENU_SHELL (menu), w);
 		gtk_widget_show (GTK_WIDGET (w));
 		g_signal_connect (G_OBJECT (w), "activate",
@@ -169,10 +169,13 @@ greeter_item_create_canvas_item (GreeterItemInfo *item)
 {
   GnomeCanvasGroup *group;
   GtkWidget *entry;
+  GtkWidget *list;
+  GtkWidget *swin;
   double x1, y1, x2, y2;
   int i;
   GtkAllocation rect;
   char *text;
+  GtkTooltips *tooltips;
 
   if (item->item != NULL)
     return;
@@ -216,6 +219,7 @@ greeter_item_create_canvas_item (GreeterItemInfo *item)
 					NULL);
     break;
   case GREETER_ITEM_TYPE_SVG:
+    gnome_i18n_push_c_numeric_locale ();
     for (i = 0; i < GREETER_ITEM_STATE_MAX; i++)
       {
 	if (item->files[i])
@@ -226,6 +230,7 @@ greeter_item_create_canvas_item (GreeterItemInfo *item)
 	else
 	  item->orig_pixbufs[i] = NULL;
       }
+    gnome_i18n_pop_c_numeric_locale ();
 
     /* Fall through */
   case GREETER_ITEM_TYPE_PIXMAP:
@@ -296,7 +301,10 @@ greeter_item_create_canvas_item (GreeterItemInfo *item)
       gtk_entry_set_invisible_char (GTK_ENTRY (entry), 0x25cf);
     
     if (item->id != NULL && strcmp (item->id, "user-pw-entry") == 0) {
-	    /* HACK! Add a menubar, this is kind of evil isn't it */
+	    /* HACK! Add a menubar, this is kind of evil isn't it,
+	     * should probably be done in the pam item setup thingie.
+	     * but this is really widget kind of thing.  I dunno where
+	     * this belongs but it's a hack here. */
 	    GtkWidget *menubar = make_menubar ();
 
 	    gnome_canvas_item_new (gnome_canvas_root (GNOME_CANVAS (canvas)),
@@ -307,6 +315,15 @@ greeter_item_create_canvas_item (GreeterItemInfo *item)
 				   "height", (double)rect.height,
 				   "width", (double)rect.width,
 				   NULL);
+
+	    /* Here add a tooltip, so that the user knows about F10 */
+	    tooltips = gtk_tooltips_new ();
+	    gtk_tooltips_set_tip (tooltips, GTK_WIDGET (entry),
+				  _("Answer questions here and press Enter "
+				    "when done.  For a menu press F10."),
+				  NULL);
+
+	    /* FIXME: how to make this accessible??? */
     }
 
     item->item = gnome_canvas_item_new (group,
@@ -319,6 +336,28 @@ greeter_item_create_canvas_item (GreeterItemInfo *item)
 					NULL);
 
     break;
+
+  case GREETER_ITEM_TYPE_LIST:
+    /* Note a list type must be setup later and we will add the list store
+     * to it then, depending on the type.  Likely userlist is the
+     * only type we support */
+    list = gtk_tree_view_new ();
+    swin = gtk_scrolled_window_new (NULL, NULL);
+    gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (swin),
+    					GTK_SHADOW_NONE);
+    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (swin),
+    				GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
+    gtk_container_add (GTK_CONTAINER (swin), list);
+    item->item = gnome_canvas_item_new (group,
+					GNOME_TYPE_CANVAS_WIDGET,
+					"widget", swin,
+					"x", x1,
+					"y", y1,
+					"height", (double)rect.height,
+					"width", (double)rect.width,
+					NULL);
+
+    break;
   }
 
   if (item->item_type == GREETER_ITEM_TYPE_RECT ||
