@@ -6,6 +6,7 @@
 #include <stdlib.h>
 #include <time.h>
 #include <unistd.h>
+#include <sys/utsname.h>
 
 #include "greeter_item.h"
 #include "greeter_configuration.h"
@@ -59,7 +60,6 @@ greeter_item_info_free (GreeterItemInfo *info)
   g_free (info->id);
   g_free (info->orig_text);
   g_free (info->show_type);
-  g_free (info->show_subtype);
   g_free (info);
 }
 
@@ -111,6 +111,7 @@ greeter_item_expand_text (const char *text)
   int r;
   int underline = -1;
   char buf[256];
+  struct utsname name;
 
   str = g_string_sized_new (strlen (text));
 
@@ -126,6 +127,10 @@ greeter_item_expand_text (const char *text)
 	    case '%':
 	      g_string_append (str, "%");
 	      break;
+	    case 'n':
+	      uname (&name);
+	      g_string_append (str, name.nodename);
+	      break;
 	    case 'h':
 	      buf[sizeof(buf) - 1] = '\0';
 	      r = gethostname (buf, sizeof(buf) - 1);
@@ -216,27 +221,32 @@ greeter_item_is_visible (GreeterItemInfo *info)
       ! (info->show_modes & GREETER_ITEM_SHOW_REMOTE))
     return FALSE;
 
-  if ( ! GdmConfigAvailable &&
+  if (( ! GdmConfigAvailable || ! GdmSystemMenu) &&
       info->show_type != NULL &&
       strcmp (info->show_type, "config") == 0)
 	  return FALSE;
 
+  if (( ! GdmChooserButton || ! GdmSystemMenu) &&
+      info->show_type != NULL &&
+      strcmp (info->show_type, "chooser") == 0)
+	  return FALSE;
+
   if ( ! GdmSystemMenu &&
       info->show_type != NULL &&
       strcmp (info->show_type, "system") == 0)
 	  return FALSE;
 
-  if (GdmHalt == NULL &&
-      info->show_subtype != NULL &&
-      strcmp (info->show_subtype, "halt") == 0)
+  if (( ! GdmSystemMenu || GdmHalt == NULL) &&
+      info->show_type != NULL &&
+      strcmp (info->show_type, "halt") == 0)
 	  return FALSE;
-  if (GdmReboot == NULL &&
-      info->show_subtype != NULL &&
-      strcmp (info->show_subtype, "reboot") == 0)
+  if (( ! GdmSystemMenu || GdmReboot == NULL) &&
+      info->show_type != NULL &&
+      strcmp (info->show_type, "reboot") == 0)
 	  return FALSE;
-  if (GdmSuspend == NULL &&
-      info->show_subtype != NULL &&
-      strcmp (info->show_subtype, "suspend") == 0)
+  if (( ! GdmSystemMenu || GdmSuspend == NULL) &&
+      info->show_type != NULL &&
+      strcmp (info->show_type, "suspend") == 0)
 	  return FALSE;
 
   if (ve_string_empty (GdmTimedLogin) &&
